
#include "StdAfx.h"
#include "WarFileDriverFileWin32.h"   // class implemented

#ifndef WAR_FILE_DRIVER_WIN32_H
#   include "WarFileDriverWin32.h"
#endif

#ifndef WAR_OS_H
#   include "WarOs.h"
#endif

#ifndef WAR_LOG_H
#   include "WarLog.h"
#endif

#ifndef WAR_FILE_H
#   include "WarFile.h"
#endif

#ifndef WAR_SVR_WIN32_NT_ENGINE_H
#   include "WarSvrWin32NtEngine.h"
#endif

#ifndef WAR_PERFMON_DEF_H
#   include "WarPerfmonDef.h"
#endif

#ifndef WAR_FUNCTION_PROFILER_H
#   include "WarFunctionProfiler.h"
#endif


#define Win32FileDriver() (*(WarFileDriverWin32 *)mpDriver)

#ifdef _DEBUG
#define Validate()\
if (mFileHandle == INVALID_HANDLE_VALUE) \
    WarThrow(WarError(WAR_ERR_NOT_OPEN), NULL);
#else
#define Validate()
#endif


/////////////////////////////// PUBLIC ///////////////////////////////////////


WarFileDriverFileWin32::ol_result_t *WarFileDriverFileWin32::mspBufferList;
WarCriticalSection WarFileDriverFileWin32::msBufferLock;

//============================= LIFECYCLE ====================================

WarFileDriverFileWin32::WarFileDriverFileWin32(WarFileDriver *pDriver)
: WarFileDriverFile(pDriver),
mFileHandle(INVALID_HANDLE_VALUE),
mCurrentFileOffset(0)
{}


WarFileDriverFileWin32::~WarFileDriverFileWin32()
{
}


//============================= OPERATORS ====================================



//============================= OPERATIONS ===================================


void
WarFileDriverFileWin32::Open(const WarUrl& openUrl, 
                             war_uint32_t openFlags) 
                             throw(WarException)
{
    if (mFileHandle != INVALID_HANDLE_VALUE)
        WarThrow(WarError(WAR_ERR_ALREADY_OPEN), NULL);

    WarLog debug_log(WARLOG_DEBUG, "WarFileDriverFileWin32::Open()");
    
    mPath = openUrl.GetFilePath();

    war_uint32_t file_access_mode = 0;
    war_uint32_t file_share_mode = 0;
    war_uint32_t file_creation_flags = 0;
    war_uint32_t file_creation_attributes = 0;
    war_uint32_t file_attributes = ::GetFileAttributes(mPath.GetPath());
    bool file_exist = (file_attributes != 0xffffffff);
    
    if (file_exist)
    {
        // Must be a file...
        if (file_attributes & FILE_ATTRIBUTE_DIRECTORY)
            WarThrow(WarError(WAR_FERR_PATH_IS_A_DIR), NULL);

        if (openFlags & F_NOTEXIST)
            WarThrow(WarError(WAR_ERR_NAME_EXIST), NULL);
    }
    
    /* Too buggy! Cannot create files 
    
      if (Win32FileDriver().GetCmpMode(openUrl) == PCMP_CASE_SENSITIVE)
    {
        file_creation_attributes |= FILE_FLAG_POSIX_SEMANTICS;
        mPath.MakePosixPath();
    } */  
    
    if (openFlags & F_READ)
    {
        file_access_mode |= GENERIC_READ;
        if (!(openFlags & (F_WRITE | F_APPEND)) &&
            !(openFlags & (F_CREATE | F_NOTEXIST | F_MUSTEXIST | F_TRUNCATE)))
            openFlags |= F_MUSTEXIST;
    }

    if (openFlags & (F_WRITE | F_APPEND))
    {
        file_access_mode |= GENERIC_WRITE;
        if (openFlags & F_TEXT)
            file_access_mode |= GENERIC_READ; // Need to read in order to seek in file

        if (!(openFlags & (F_READ)) &&
            !(openFlags & (F_CREATE | F_NOTEXIST | F_MUSTEXIST | F_TRUNCATE)))
            openFlags |= F_CREATE;
    }
    
    if (openFlags & F_SHREAD)
        file_share_mode |= FILE_SHARE_READ;
    
    if (openFlags & F_SHWRITE)
        file_share_mode |= FILE_SHARE_WRITE;
    
    if (openFlags & F_EXCLUSIVE)
        file_share_mode &= ~(F_SHWRITE | F_SHREAD);
    
    if (openFlags & F_SEQUENTIAL)
        file_creation_attributes |= FILE_FLAG_SEQUENTIAL_SCAN;
    
    if ((openFlags & F_CALLBACK) && WarOs::GetOs().IsNt())
        file_creation_attributes |= FILE_FLAG_OVERLAPPED;
    
    
    switch (openFlags & (F_CREATE | F_NOTEXIST | F_MUSTEXIST | F_TRUNCATE))
    {
    case F_CREATE:
        if (openFlags & F_APPEND)
            file_creation_flags = OPEN_ALWAYS;
        else
            file_creation_flags = CREATE_ALWAYS;
        break;
    case F_CREATE | F_NOTEXIST:
        file_creation_flags = CREATE_NEW;
        break;
    case F_CREATE | F_MUSTEXIST:
        file_creation_flags = OPEN_EXISTING;
        break;
    case F_TRUNCATE:
    case F_CREATE | F_TRUNCATE:
        // TRUNCATE_EXISTING will fail if the file don't exist.
        if (!file_exist)
        {
            file_creation_flags = CREATE_ALWAYS;	// File don't exist
            break;
        }
        // Fall trough
    case F_MUSTEXIST | F_TRUNCATE:
    case F_CREATE | F_MUSTEXIST | F_TRUNCATE:
        file_creation_flags = TRUNCATE_EXISTING;
        //bCheckBannedList = TRUE;
        break;
    case F_NOTEXIST:
        file_creation_flags = CREATE_NEW;
        //bCheckBannedList = TRUE;
        break;
    case F_MUSTEXIST:
        file_creation_flags = OPEN_EXISTING;
        break;
    default:
        WarThrow(WarError(WAR_ERR_INVALID_ARGUMENT), NULL);
    }

retry_open:
    int num_retries = 0;
    
    mFileHandle = ::CreateFile(mPath.GetPath(), 
        file_access_mode, file_share_mode, NULL, 
        file_creation_flags, file_creation_attributes, NULL);
    
    if (mFileHandle == INVALID_HANDLE_VALUE)
    {
        WarSystemError syserr;

        if (!num_retries++ 
            && (syserr.SystemError() == ERROR_LOGON_FAILURE))
        {
            if (mpDriver->LogonToUrl(openUrl))
                goto retry_open; 
        }
        
        if (debug_log)
        {
            debug_log << "Open '"
                << mPath.GetPath()
                << "' failed. Open flags: "
                << syserr
                << mpDriver->WarExplainFags(openFlags)
                << war_endl;
        }
        
        WarThrow(syserr, NULL);
    }

    if (openFlags & F_CALLBACK)
    {
        if (::CreateIoCompletionPort(mFileHandle,
            WarSvrWin32NtEngine::GetEngine().GetPort(),
            (DWORD)FileIOCompletionRoutine,
            0) == NULL)
        {
            WarSystemError system_error;

            ::CloseHandle(mFileHandle);
            mFileHandle = INVALID_HANDLE_VALUE;
            
            if (debug_log)
            {
                debug_log << "Open '"
                    << mPath.GetPath()
                    << "' failed on CreateIoCompletionPort. openFlags: "
                    << system_error
                    << mpDriver->WarExplainFags(openFlags)
                    << war_endl;
            }
            
            WarThrow(system_error, "CreateIoCompletionPort()");
        }
    }
    
    if (openFlags & F_APPEND)
        Seek(0, WAR_SEEK_END);
    
    //done:
    mFlags = openFlags;
}


void
WarFileDriverFileWin32::Close() throw(WarException)
{
    if (mFileHandle != INVALID_HANDLE_VALUE)
        ::CloseHandle(mFileHandle);
    mFileHandle = INVALID_HANDLE_VALUE;
}

war_uint32_t
WarFileDriverFileWin32::Read(war_cptr_t Buf, 
                             war_uint32_t Bytes)
                             throw(WarException)
{
    Validate();
    
    DWORD dwBytesRead = 0;
    if (!::ReadFile(mFileHandle, Buf, Bytes, &dwBytesRead, NULL))
    {
        WarError syserr;
        syserr.Capture();

        if (syserr.SystemError() == ERROR_HANDLE_EOF)
            syserr = WAR_FERR_END_OF_FILE;
        
        WarThrow(syserr, NULL);
    }
    
    mCurrentFileOffset += dwBytesRead;
    return dwBytesRead;
}

void
WarFileDriverFileWin32::Write(war_ccptr_t Buf, 
                              war_uint32_t Bytes)
                              throw(WarException)
{
    Validate();
    DWORD dwBytesWritten = 0;
    if (!::WriteFile(mFileHandle, Buf, Bytes, &dwBytesWritten, NULL))
    {
        WarError sys_err;
        sys_err.Capture(WAR_FERR_WRITE_FAILED);
        WarThrow(sys_err, NULL);
    }

    if (Bytes != dwBytesWritten)
    {
        // Seek back to preserve file-offset integrity
        if (dwBytesWritten)
            Seek(mCurrentFileOffset - dwBytesWritten, WAR_SEEK_BEGIN);
        
        WarThrow(WarError(WAR_FERR_WRITE_FAILED), NULL);
    }

    mCurrentFileOffset += dwBytesWritten;
}

void
WarFileDriverFileWin32::Flush() throw(WarException)
{
    Validate();
    ::FlushFileBuffers(mFileHandle);
}

war_flen_t WarFileDriverFileWin32::GetLength() throw(WarException)
{
    Validate();

    DWORD dwSizeLow = 0, dwSizeHigh = 0;

    dwSizeLow = GetFileSize(mFileHandle, &dwSizeHigh);
    if (dwSizeLow == 0xffffffff)
    {
        WarError system_error;
        system_error.Capture();
        if (system_error.SystemError() != NO_ERROR)
            WarThrow(system_error, NULL);
    }
    
    LARGE_INTEGER li;
    li.LowPart = dwSizeLow;
    li.HighPart = dwSizeHigh;

    return li.QuadPart;
}

war_flen_t
WarFileDriverFileWin32::Seek(war_flen_t fileOffset, 
                             SeekModes seekMode)
                             throw(WarException)
{
    Validate();

    war_flen_t new_offset = mCurrentFileOffset;

    switch(seekMode)
    {
    case WAR_SEEK_BEGIN:
        new_offset = fileOffset;
        break;
    case WAR_SEEK_CURRENT:
        new_offset += fileOffset;
        break;
    case WAR_SEEK_END:
        new_offset = GetLength() + fileOffset;
        break;
    default:
        WarThrow(WarError(WAR_ERR_INVALID_ARGUMENT), NULL);
    }

    if (new_offset < 0)
        WarThrow(WarError(WAR_ERR_INVALID_ARGUMENT), NULL);

    LARGE_INTEGER li;
    
    li.QuadPart = new_offset;

    li.LowPart = ::SetFilePointer(mFileHandle, li.LowPart, 
        &li.HighPart, FILE_BEGIN);
    
    WarError sys_err;
    sys_err.Capture();
    if (li.LowPart == 0xFFFFFFFF 
        && (sys_err.SystemError() != NO_ERROR))
        WarThrow(sys_err, NULL);
    
    return mCurrentFileOffset = li.QuadPart;
}

void WarFileDriverFileWin32::WriteWithCallback(
    war_transfer_buffer_ptr_t& outBuffer)
    throw(WarException)
{
    WAR_PROFILE_FUNC(WAR_FP_WarFileDriverFileWin32_WriteWithCallback);

    DWORD dwBytes = 0;
    ol_result_t *pResult = AllocResult(OP_WRITE);
    pResult->mBufferPtr = outBuffer;

    LARGE_INTEGER li;
    li.QuadPart = mCurrentFileOffset;
    pResult->mOverlappedData.Offset = li.LowPart;
    pResult->mOverlappedData.OffsetHigh  = li.HighPart ;

    if (!WriteFile(mFileHandle, 
        pResult->mBufferPtr->GetBuffer(),
        pResult->mBufferPtr->mBytesUsed,
        &dwBytes,
        &pResult->mOverlappedData))
    {
        WarSystemError system_error;
        
        switch(system_error.SystemError())
        {
        case ERROR_IO_PENDING:
            break; // No action
        case ERROR_HANDLE_EOF:
            DeleteResult(pResult);
            WarThrow(WarError(WAR_FERR_END_OF_FILE, ERROR_HANDLE_EOF), NULL);
            break;
        default:
            DeleteResult(pResult);
            WarThrow(system_error, NULL);
        }
    }

    // May move the internal file-offset off the physical size!
    mCurrentFileOffset += outBuffer->GetLength();
    
    WAR_DB_PERFMON_INC(WAR_PRFDEBUG_NUM_FILEPENDINGIO);
}

void WarFileDriverFileWin32::ReadWithCallback(
    war_transfer_buffer_ptr_t& inBuffer)
    throw(WarException)
{
    WAR_PROFILE_FUNC(WAR_FP_WarFileDriverFileWin32_ReadWithCallback);


    DWORD dwBytes = 0;
    ol_result_t *pResult;

    {
        WAR_PROFILE_FUNC(WAR_FP_WarFileDriverFileWin32_ReadWithCallback_ALLOC);
        pResult = AllocResult(OP_READ);
    }

    {
        WAR_PROFILE_FUNC(WAR_FP_WarFileDriverFileWin32_ReadWithCallback_PREP);
        
        pResult->mBufferPtr = inBuffer;

        LARGE_INTEGER li;
        li.QuadPart = mCurrentFileOffset;
        pResult->mOverlappedData.Offset = li.LowPart;
        pResult->mOverlappedData.OffsetHigh  = li.HighPart ;
    }

    BOOL result;
    
    {
        WAR_PROFILE_FUNC(WAR_FP_WarFileDriverFileWin32_ReadWithCallback_READFILE);

        result = ReadFile(mFileHandle, 
            pResult->mBufferPtr->GetBuffer(),
            pResult->mBufferPtr->GetLength(),
            &dwBytes,
            &pResult->mOverlappedData);
    }

    {
        WAR_PROFILE_FUNC(WAR_FP_WarFileDriverFileWin32_ReadWithCallback_POSTPRC);
        
        if (!result)
        {
            WarSystemError system_error;
            WarError throw_error;
            
            switch(system_error.SystemError())
            {
            case ERROR_IO_PENDING:
                break; // No action
            case ERROR_HANDLE_EOF:
                throw_error = WarError(WAR_FERR_END_OF_FILE, ERROR_HANDLE_EOF);
                break;
            default:
                throw_error = system_error;
                break;
            }
            
            if (throw_error.LocalError() != WAR_ERR_OK)
            {
                DeleteResult(pResult);
                WarThrow(throw_error, NULL);
            }
        }
        
        // May move the internal file-offset off the physical size!
        mCurrentFileOffset += inBuffer->GetLength();
    }

    WAR_DB_PERFMON_INC(WAR_PRFDEBUG_NUM_FILEPENDINGIO);
}

//============================= ACCESS     ===================================
//============================= INQUIRY    ===================================

bool WarFileDriverFileWin32::IsOpen()
{
    return (mFileHandle != INVALID_HANDLE_VALUE);
}


/////////////////////////////// PROTECTED  ///////////////////////////////////

WarFileDriverFileWin32::ol_result_def *
WarFileDriverFileWin32::AllocResult(WarFileDriverFileWin32::IoOperationE operation)
{
    WarAutoLock my_lock(msBufferLock);

    ol_result_t *p;

    if (mspBufferList)
    {
        p = mspBufferList;
        mspBufferList = mspBufferList->mpNext;
    }
    else
    {
        WAR_PROFILE_FUNC(WAR_FP_WarFileDriverFileWin32_AllocResult_NEW);

        p = new ol_result_t;
    }
    
    
    WAR_PROFILE_FUNC(WAR_FP_WarFileDriverFileWin32_AllocResult_ASSIGN);
    
    p->mpThis = this;
    p->mFilePtr = mpCompanion;
    memset(&(p->mOverlappedData), 0, sizeof(OVERLAPPED));
    p->mOperation = operation;
    
    
    return p;
}


VOID CALLBACK WarFileDriverFileWin32::FileIOCompletionRoutine(
        DWORD dwErrorCode,                // completion code
        DWORD dwNumberOfBytesTransfered,  // number of bytes transferred
        LPOVERLAPPED lpOverlapped         // I/O information buffer
        )
{
    WarLog err_log(WARLOG_ERROR, "WarFileDriverFileWin32::FileIOCompletionRoutine()");
    ol_result_t *pResult = (ol_result_def *)lpOverlapped;
    WarError err((dwErrorCode != WAR_ERR_OK) 
        ? WAR_ERR_SYSTEM_ERROR : WAR_ERR_OK, (int)dwErrorCode);

    try
    {
        pResult->mBufferPtr->mBytesUsed 
            = dwNumberOfBytesTransfered;

        if (err && (err.SystemError() == ERROR_HANDLE_EOF))
            err = WAR_FERR_END_OF_FILE;

        switch(pResult->mOperation)
        {
        case OP_READ:
            WAR_DB_PERFMON_DEC(WAR_PRFDEBUG_NUM_FILEPENDINGIO);
            pResult->mpThis->OnRead(err, pResult->mBufferPtr);
            break;
        case OP_WRITE:
            WAR_DB_PERFMON_DEC(WAR_PRFDEBUG_NUM_FILEPENDINGIO);
            pResult->mpThis->OnWritten(err, pResult->mBufferPtr);
            break;
        }
    }
    
    catch(WarException& ex)
    {
        err_log << "Caught unexcpected exception. Operation was "
            << pResult->mOperation
            << ' '
            << ex.Explain();
            war_endl;
    }
#if WAR_CATCH_ALL
    catch(...)
    {
         err_log << "Caught unhandled exception. Operation was "
            << pResult->mOperation
            << war_endl;
    }
#endif
    
    DeleteResult(pResult);
}

void WarFileDriverFileWin32::DeleteResult(ol_result_def *presult)
{
    WarAutoLock my_lock(msBufferLock);

    // Prepere for re-use
    presult->mpNext = mspBufferList;
    mspBufferList = presult;

    // Release autopointers
    presult->mFilePtr = NULL;
    presult->mBufferPtr = NULL;
}


/////////////////////////////// PRIVATE    ///////////////////////////////////

