#include "StdAfx.h"
#include "WarWin32Registry.h"   // class implemented

#ifndef WAR_INCLUDED_SHLWAPI_H
#   define WAR_INCLUDED_SHLWAPI_h
#   include "shlwapi.h"
#endif

#ifndef WAR_INCLUDED_TCHAR_H
#   define WAR_INCLUDED_TCHAR_H
#   include <tchar.h>
#endif

#if WAR_CRYPTO && !defined(WAR_OPENSSL_RSA_H_INCLUDED)
#   define WAR_OPENSSL_RSA_H_INCLUDED
#   include <openssl/rsa.h>
#endif

#if WAR_CRYPTO && !defined(WAR_OPENSSL_PEM_H_INCLUDED)
#   define WAR_OPENSSL_PEM_H_INCLUDED
#   include <openssl/pem.h>
#endif

#ifndef WAR_LOG_H
#   include "WarLog.h"
#endif

#define DEFAULT_REGKEY_BUFLEN (1024 * 4 * sizeof(TCHAR))
#define DEFAULT_NAME_BUFFER_LEN (512)

using namespace std;

/////////////////////////////// PUBLIC ///////////////////////////////////////

war_registrypath_t WarWin32Registry::mRootPath;

#define BL_LEN EVP_CIPHER_block_size(ctx.cipher)
#define BL_ALIGN(a) a += (BL_LEN - (a % BL_LEN))


//============================= LIFECYCLE ====================================

WarWin32Registry::WarWin32Registry()
: mNodeKey(0),
mRootKey(0)
{
}

WarWin32Registry::~WarWin32Registry()
{
    Reset();
}

//============================= OPERATORS ====================================
//============================= OPERATIONS ===================================

void WarWin32Registry::Open(const open_t& openVal, 
                            int regMode)
                            throw(WarException)
{
    mCrypto = openVal.GetReg().mCrypto;
    Open(openVal.GetReg().mRootKey, openVal.GetPath());
}

void WarWin32Registry::Open(HKEY keyRoot, 
                              const war_registrypath_t& regPath, 
                              int regMode)
                              throw(WarException)
{
    LONG status = 0;

    Reset();

    mPath = regPath.IsEmpty() ? mRootPath : regPath;

    if ((status = ::RegOpenKeyEx(
        keyRoot,
        mPath.GetPath(),
        0,
        regMode,
        &mNodeKey)) != ERROR_SUCCESS)
    {
        WarError sys_err(WAR_ERR_SYSTEM_ERROR, status);
        if (status == ERROR_FILE_NOT_FOUND)
            sys_err = WAR_ERR_NAME_DONT_EXIST;
        WarThrow(sys_err, NULL);
    }

    mRootKey = keyRoot;
}

void WarWin32Registry::Create(const open_t& openVal, 
                            bool doFailIfNotExist)
                            throw(WarException)
{
    mCrypto = openVal.GetReg().mCrypto;
    Create(openVal.GetReg().mRootKey, 
        openVal.GetPath(), doFailIfNotExist);
}

void WarWin32Registry::Create(HKEY keyRoot, 
        const war_registrypath_t& regPath,
        bool doFailIfNotExist)
        throw(WarException)
{
    DWORD key_disposition = 0;
    LONG status = 0;

    mPath = regPath.IsEmpty() ? mRootPath : regPath;

    Reset();

    if ((status = ::RegCreateKeyEx(
        keyRoot,
        mPath.GetPath(),
        0,
        NULL,
        REG_OPTION_NON_VOLATILE,
        KEY_WRITE | KEY_READ,
        NULL,
        &mNodeKey,
        &key_disposition)) != ERROR_SUCCESS)
    {
        WarThrow(WarSystemError(), NULL);
    }

    if (doFailIfNotExist 
        && (REG_OPENED_EXISTING_KEY == key_disposition))
    {
        Reset();
        WarThrow(WarError(WAR_ERR_OBJECT_EXIST), NULL);
    }

    mRootKey = keyRoot;
}

war_regstr_t 
WarWin32Registry::GetStrValue(war_registrypath_ccstr_t valueName,
                           war_registrypath_ccstr_t defaultValue,
                           bool doFailIfNotExist,
                           bool doDecrypt) const
                           throw(WarException)
{
    war_regstr_t return_val;
    LONG status;
    char byte_buffer[DEFAULT_REGKEY_BUFLEN];
    DWORD dw_len = sizeof(byte_buffer), dw_type = 0;

#if WAR_CRYPTO
    if (doDecrypt)
    {
        if (GetEncryptedStrValue(valueName, return_val))
            return return_val;
    }
#endif

    if ((status = ::RegQueryValueEx(mNodeKey, valueName, 
        0, &dw_type, (LPBYTE)byte_buffer, &dw_len)) != ERROR_SUCCESS)
    {
        if((ERROR_FILE_NOT_FOUND == status)
            && !doFailIfNotExist)
        {
            return defaultValue ? defaultValue : _T("");
        }
        WarThrow(WarError(WAR_ERR_SYSTEM_ERROR, status), NULL);
    }

    switch(dw_type)
    {
        //case REG_DWORD:
        case REG_DWORD_LITTLE_ENDIAN:
        case REG_DWORD_BIG_ENDIAN:
            {
                DWORD my_dw_value = *(DWORD *)byte_buffer;
                if (REG_DWORD_BIG_ENDIAN == dw_type)
                    my_dw_value = ::ntohl(my_dw_value);
                WarItoa(return_val, my_dw_value);
            }
            break;
        case REG_EXPAND_SZ:
            {
                char tmp_buffer[DEFAULT_REGKEY_BUFLEN];
                if (!ExpandEnvironmentStrings(
                    (TCHAR *)byte_buffer, 
                    (TCHAR *)tmp_buffer, 
                    DEFAULT_REGKEY_BUFLEN / sizeof(TCHAR)) )
                {
                    WarThrow(WarSystemError(), NULL);
                }
                return_val = (TCHAR *)tmp_buffer;
            }
            break;
        case REG_SZ:
            return_val = (TCHAR *)byte_buffer;
            break;
        case REG_NONE:
            break;
#if 0  //  REG_QWORD
        case REG_QWORD:
        //case REG_QWORD_LITTLE_ENDIAN:
            {
                QWORD my_dw_value = *(QWORD *)byte_buffer;
                WarItoa(return_val, my_dw_value);
            }
            break;
#endif // REG_QWORD
        default:
            WarThrow(WarError(WAR_ERR_INVALID_CASE_VALUE), NULL);
            break;
    }

    return return_val;
}

DWORD WarWin32Registry::GetIntValue(war_registrypath_ccstr_t valueName,
                               DWORD defaultValue,
                               bool doFailIfNotExist) const
                               throw(WarException)
{
    DWORD return_val;
    LONG status;
    char byte_buffer[DEFAULT_REGKEY_BUFLEN];
    DWORD dw_len = sizeof(byte_buffer), dw_type = 0;

    if ((status = ::RegQueryValueEx(mNodeKey, valueName, 
        0, &dw_type, (LPBYTE)byte_buffer, &dw_len)) != ERROR_SUCCESS)
    {
        if (doFailIfNotExist)
            WarThrow(WarError(WAR_ERR_SYSTEM_ERROR, status), NULL);
        return defaultValue;
    }

    switch(dw_type)
    {
        //case REG_DWORD:
        case REG_DWORD_LITTLE_ENDIAN:
        case REG_DWORD_BIG_ENDIAN:
            {
                DWORD my_dw_value = *(DWORD *)byte_buffer;
                if (REG_DWORD_BIG_ENDIAN == dw_type)
                    my_dw_value = ::ntohl(my_dw_value);
                return_val = my_dw_value;
            }
            break;
        default:
            WarThrow(WarError(WAR_ERR_INVALID_CASE_VALUE), NULL);
            break;
    }

    return return_val;
}

WarWin32Registry::regstr_list_t 
WarWin32Registry::GetStrListValue(war_registrypath_ccstr_t valueName,
        regstr_list_t& defaultValue,
        bool doFailIfNotExist) const 
        throw(WarException)
{
    regstr_list_t return_val;
    war_regstr_t str_val;
    LONG status;
    char byte_buffer[DEFAULT_REGKEY_BUFLEN];
    DWORD dw_len = sizeof(byte_buffer), dw_type = 0;

    if ((status = ::RegQueryValueEx(mNodeKey, valueName, 
        0, &dw_type, (LPBYTE)byte_buffer, &dw_len)) != ERROR_SUCCESS)
    {
        WarThrow(WarError(WAR_ERR_SYSTEM_ERROR, status), NULL);
    }

    switch(dw_type)
    {
        //case REG_DWORD:
        case REG_DWORD_LITTLE_ENDIAN:
        case REG_DWORD_BIG_ENDIAN:
            {
                DWORD my_dw_value = *(DWORD *)byte_buffer;
                if (REG_DWORD_BIG_ENDIAN == dw_type)
                    my_dw_value = ::ntohl(my_dw_value);
                WarItoa(str_val, my_dw_value);
                return_val.push_back(str_val);
            }
            break;
        case REG_EXPAND_SZ:
            {
                char tmp_buffer[DEFAULT_REGKEY_BUFLEN];
                if (!ExpandEnvironmentStrings(
                    (TCHAR *)byte_buffer, 
                    (TCHAR *)tmp_buffer, 
                    DEFAULT_REGKEY_BUFLEN / sizeof(TCHAR)) )
                {
                    WarThrow(WarSystemError(), NULL);
                }
                str_val = (TCHAR *)tmp_buffer;
                return_val.push_back(str_val);
            }
            break;
        case REG_SZ:
            str_val = (TCHAR *)byte_buffer;
            return_val.push_back(str_val);
            break;
        case REG_MULTI_SZ:
            {
                for(TCHAR *p = (TCHAR *)byte_buffer; *p; *p++)
                {
                    return_val.push_back(war_regstr_t(p));
                    p += return_val.back().size();
                }
            }
            break;
        case REG_NONE:
            break;
#if defined(REG_QWORD) && defined(HAVE_QWORD)
        case REG_QWORD:
        //case REG_QWORD_LITTLE_ENDIAN:
            {
                QWORD my_dw_value = *(QWORD *)byte_buffer;
                WarItoa(str_val, my_dw_value);
                return_val.push_back(str_val);
            }
            break;
#endif // REG_QWORD
        default:
            WarThrow(WarError(WAR_ERR_INVALID_CASE_VALUE), NULL);
            break;
    }

    return return_val;
}

static char *last_encr_buf;

void WarWin32Registry::SetValue(war_registrypath_ccstr_t valueName, 
                                const war_regstr_t& newValue,
                                bool doEncrypt) const
                                throw(WarException)
{
    LONG status = 0;
    CONST BYTE * new_value;
    size_t bytes = 0;
    
#if WAR_CRYPTO
    war_registrypath_t cr_path;
    war_regstr_t encr_value;

    if (doEncrypt)
    {
        // Delete the unencrypted version, if it exists
        try
        {
            DeleteValue(valueName);
        }
        catch(WarException& ex)
        {
            if (ex.LocalError() != WAR_ERR_OBJECT_NOT_FOUND)
                throw ex;
        }

        // Remap the name to <keyname>.cr
        // That way we can recognize encrypted values.
        cr_path << valueName << _T(".cr");
        valueName = cr_path.GetPath();

        mCrypto.Encrypt(encr_value, newValue);

        new_value = (CONST BYTE *)(encr_value.c_str());
        bytes = (encr_value.size() +1) * sizeof(TCHAR);
    }
    else
#endif // WAR_CRYPTO
    {
        new_value = (CONST BYTE *)(newValue.c_str());
        bytes = (newValue.size() +1) * sizeof(TCHAR);
    }

    if ((status = ::RegSetValueEx(mNodeKey, valueName, 0,
        REG_SZ, new_value, bytes)) != ERROR_SUCCESS)
    {
        WarThrow(WarError(WAR_ERR_SYSTEM_ERROR, status), NULL);
    }
}


void WarWin32Registry::SetValue(war_registrypath_ccstr_t valueName, 
                                DWORD newValue) const
                                throw(WarException)
{
    LONG status = 0;

    if ((status = ::RegSetValueEx(mNodeKey, valueName, 0,
        REG_DWORD, (CONST BYTE *)(&newValue),
        sizeof(DWORD))) != ERROR_SUCCESS)
    {
        WarThrow(WarError(WAR_ERR_SYSTEM_ERROR, status), NULL);
    }
}


void WarWin32Registry::DeleteKey(war_registrypath_ccstr_t keyName) 
{
    if (keyName == NULL)
    {
        WarWin32Registry my_key;
        Reset();

        my_key.Open(mRootKey, mPath.GetPathname().GetPath(), KEY_WRITE);
        
        my_key.DeleteKey(mPath.GetFilename().GetPath());
        return;
    }

    LONG status = ::SHDeleteKey(mNodeKey, keyName);
    if (status != ERROR_SUCCESS)
        WarThrow(WarError(WAR_ERR_SYSTEM_ERROR, status), NULL);
}


void WarWin32Registry::DeleteValue(war_registrypath_ccstr_t valueName) const
{
    LONG status = ::RegDeleteValue(mNodeKey, valueName);
    if (status != ERROR_SUCCESS)
    {
        switch(status)
        {
        case ERROR_FILE_NOT_FOUND:
            WarThrow(WarError(WAR_ERR_OBJECT_NOT_FOUND, status), NULL);
        default:
            WarThrow(WarError(WAR_ERR_SYSTEM_ERROR, status), NULL);
        }
    }
}


void WarWin32Registry::EnumKey(keyname_list_t& resultList)
{
    int indedx = 0;
    LONG status;
    TCHAR name_buffer[DEFAULT_NAME_BUFFER_LEN];
    DWORD name_len = DEFAULT_NAME_BUFFER_LEN, class_len = 0;

    while((status = ::RegEnumKeyEx(mNodeKey,
        indedx++,
        name_buffer,
        &name_len,
        NULL,
        NULL,
        &class_len,
        NULL)) == ERROR_SUCCESS)
    {
        resultList.push_back(war_regstr_t(name_buffer));
        name_len = DEFAULT_NAME_BUFFER_LEN;
        class_len = 0;
    }

    if (ERROR_NO_MORE_ITEMS != status)
        WarThrow(WarError(WAR_ERR_SYSTEM_ERROR, status), NULL);
}


void WarWin32Registry::EnumValue(keyname_list_t& resultList)
{
    int indedx = 0;
    LONG status;
    TCHAR name_buffer[DEFAULT_NAME_BUFFER_LEN];
    DWORD name_len = DEFAULT_NAME_BUFFER_LEN;

    while((status = ::RegEnumValue(mNodeKey,
        indedx++,
        name_buffer,
        &name_len,
        NULL,
        NULL,
        NULL,
        NULL)) == ERROR_SUCCESS)
    {
        resultList.push_back(war_regstr_t(name_buffer));
        name_len = DEFAULT_NAME_BUFFER_LEN;
    }

    if (ERROR_NO_MORE_ITEMS != status)
        WarThrow(WarError(WAR_ERR_SYSTEM_ERROR, status), NULL);
}


void WarWin32Registry::Flush()
{
}

void WarWin32Registry::Rename(war_registrypath_ccstr_t newName)
        throw(WarException)
{
    if (!newName || !*newName)
        WarThrow(WarError(WAR_ERR_INVALID_ARGUMENT), NULL);

    war_registrypath_t new_path = mPath.GetPathname().GetPath();

    new_path << WAR_SYSSLASH << newName;

    try
    {
        WarWin32Registry test_key;
        test_key.Open(mRootKey, new_path, KEY_READ);
        WarThrow(WarError(WAR_ERR_NAME_EXIST), NULL);
    }
    catch(WarException)
    {
    }

    WarWin32Registry new_key, father_key;

    father_key.Open(mRootKey, 
        mPath.GetPathname().GetPath(), 
        KEY_ALL_ACCESS);
    
    new_key.Create(mRootKey, new_path);

    DWORD status = SHCopyKey(father_key.mNodeKey, 
        mPath.GetFilename().GetPath(),
        new_key.mNodeKey,
        NULL);
    if (status != ERROR_SUCCESS)
    {
        new_key.DeleteKey();
        WarThrow(WarError(WAR_ERR_SYSTEM_ERROR, status), NULL);
    }

    DeleteKey();

    // Steal the new properties from new_key
    mNodeKey = new_key.mNodeKey;
    new_key.mNodeKey = NULL;
    mRootKey = new_key.mRootKey;
    new_key.mRootKey = NULL;
    mPath = new_key.mPath;
}

#if WAR_CRYPTO
void WarWin32Registry::LoadPamFile() throw(WarException)
{
    WarLog err_log(WARLOG_ERROR, "WarWin32Registry::LoadPamFile()");

    war_syspath_t key_file = GetStrValue(
        WAR_WINNT_REG_PKEYFILE, NULL, true).c_str();

    FILE *fp = _tfopen(key_file.GetPath(), _T("rt"));
    if (NULL == fp)
    {
        WarError sys_err(WAR_ERR_SYSTEM_ERROR, errno);
        err_log << "Failed to open pam file "
            << key_file.GetPath()
            << sys_err
            << war_endl;

        WarThrow(sys_err, NULL);
    }

    EVP_PKEY *key = PEM_read_PrivateKey(
        fp, 
        NULL,
        NULL,
        NULL);

    fclose(fp);

    if (NULL == key)
    {
        WarOpenSslError sys_err;
        err_log << "PEM_read_PrivateKey() failed when reading pam file "
            << key_file.GetPath()
            << sys_err
            << war_endl;

        WarThrow(sys_err, "PEM_read_PrivateKey()");
    }

    char *strkey = BN_bn2hex(key->pkey.dh->priv_key);

    WarCryptText key_encr(strkey);
    WarCollector<TCHAR> data_detrypt_key(WarCollector<TCHAR>::SM_ERASE);
    key_encr.Decrypt(data_detrypt_key.mString,
        GetStrValue(WAR_WINNT_REG_CRYPTKEY, NULL, true));

    mCrypto.SetPassword(WarCollector<char>(data_detrypt_key).GetValue().c_str());
    
    memset(strkey, 0, strlen(strkey));
    OPENSSL_free(strkey);
    EVP_PKEY_free(key);
}
#endif WAR_CRYPTO


//============================= ACCESS     ===================================
//============================= INQUIRY    ===================================

const WarWin32Registry::open_t 
WarWin32Registry::GetRef(war_registrypath_ccstr_t SubKeyName) const
{
    war_registrypath_t my_path = mPath;
    if (SubKeyName)
        my_path << WAR_SYSSLASH << SubKeyName;

    return open_t(*this, my_path);
}

bool WarWin32Registry::HaveValue(war_registrypath_ccstr_t valueName)
const
{
    DWORD dw_type = 0, dw_len = 0;
    LONG status = 0;

    if ((status = ::RegQueryValueEx(mNodeKey, valueName, 
        0, &dw_type, NULL, &dw_len)) != ERROR_SUCCESS)
    {
        if(ERROR_FILE_NOT_FOUND == status)
            return false;
    }

    return true;
}

bool WarWin32Registry::HaveKey(war_registrypath_ccstr_t keyName)
const
{
    try
    {
        WarWin32Registry tmp_key;
        tmp_key.Open(GetRef(keyName));
    }
    catch(WarException)
    {
        return false;
    }

    return true;
}

/////////////////////////////// PROTECTED  ///////////////////////////////////

/////////////////////////////// PRIVATE    ///////////////////////////////////

void WarWin32Registry::Reset()
{
    if (IsOpen())
        ::CloseHandle(mNodeKey);
    
    mNodeKey = NULL;
}

#if WAR_CRYPTO

bool WarWin32Registry::GetEncryptedStrValue(
    war_registrypath_ccstr_t valueName,
    war_regstr_t& return_val)
    const throw(WarException)
{
    war_registrypath_t cr_path;
    cr_path << valueName << _T(".cr");
    war_regstr_t hex_encoded_value;
    
    try
    {
        hex_encoded_value = GetStrValue(cr_path.GetPath(), 
            NULL, true);
    }
    catch(WarException& ex)
    {
        if (ex.SystemError() == ERROR_FILE_NOT_FOUND)
            return false; // Not encrypted
            
        throw ex;
    }

    mCrypto.Decrypt(return_val, hex_encoded_value);
   
    return true;
}

#endif // WAR_CRYPTO

