/** @name Base class for streaming input
 *
 * #include "WarCollector.h" <BR>
 *
 * @doc This class is used as a base-class as a
 * string class substitute for controlled, streaming
 * input with various datatypes
 *  
 * @see something
 */

#ifndef WAR_COLLECTOR_H
#define WAR_COLLECTOR_H

/* SYSTEM INCLUDES */

#ifndef WAR_STDLIB_H_INCLUDED
#   include <stdlib.h>
#endif

#ifndef WAR_VECTOR_INCLUDED
#   include <vector>
#endif

#ifndef WAR_STRING_INCLUDED
#   include <string>
#endif

#ifndef WAR_MALLOC_H_INCLUDED
# define WAR_MALLOC_H_INCLUDED
#   include <malloc.h>
#endif

/* PROJECT INCLUDES */

#ifndef WAR_TYPES_H
#   include "WarTypes.h"
#endif

#ifndef WAR_ITOA_H
#   include "WarItoa.h"
#endif

#ifndef WAR_ERROR_H
#   include "WarError.h"
#endif

#ifndef WAR_EXCEPTION_H
#   include "WarException.h"
#endif

/* LOCAL INCLUDES */


/* FORWARD REFERENCES */

enum war_endl_e 
{
    /** */
    war_endl = '\n'
};

#ifdef __cplusplus
extern "C" 
{
#endif
/****************** BEGIN OLD STYLE C spesific ********/


/****************** END OLD STYLE C spesific **********/
#ifdef __cplusplus 
}
#endif

/****************** BEGIN C++ spesific ****************/

#ifdef __cplusplus

template <class charT=char>
class WarCollector 
{
public:
    typedef charT type_t;

    enum SecurityModeE
    {
        SM_NONE,
        SM_ERASE
    };
    // LIFECYCLE
    
    /**
    * Default constructor.
    */
    WarCollector<charT>(const SecurityModeE securityMode = SM_NONE)
        : mSecurityMode(securityMode)
    {}
    
    /**
    * Copy constructor.
    *
    * @param from The value to copy to this object.
    */
    template <class T>
    WarCollector<charT>(const T& from,
        const SecurityModeE securityMode = SM_NONE)
        : mSecurityMode(securityMode)
    {
        *this << from;
    }
    
    /**
    * Destructor.
    */
    virtual ~WarCollector<charT>(void)
    {
        Reset();
    }
    
    // OPERATORS
    /** Assignment operator */
   
    /**
    * Assignment operator.
    *
    * @param from The value to assign to this object.
    *
    * @return A reference to this object.
    */
    template <class T>
    WarCollector<charT>& operator =
        (WarCollector<T>& from)
    {
        Reset();
        return *this << from;
    }

    /**
    * Assignment operator.
    *
    * @param from The value to assign to this object.
    *
    * @return A reference to this object.
    */
    WarCollector<charT>& operator=(std::string& from)
    {
        Reset();
        return *this << from;
    }
    

    /**
    * Assignment operator.
    *
    * @param from The value to assign to this object.
    *
    * @return A reference to this object.
    */
    WarCollector<charT>& operator=(war_ccstr_t from)
    {
        Reset();
        return *this << from;
    }

    bool operator == (const charT *strName)
    {
        return mString == std::basic_string<charT>(strName);
    }
    
    // OPERATIONS

    void ExpandNewline() 
    {
#ifdef WIN32
        char *newbuf = (charT *)alloca(
            mString.size() * 2 * sizeof(charT));
        const char *p = mString.c_str();
        char *pp = newbuf;
        bool have_nl = false;

        while(*p)
        {
            if (*p == '\n')
            {
                have_nl = true;
                *pp++ = '\r';
            }
            *pp++ = *p++;
        }

        if (have_nl)
        {
            *pp = 0;
            Reset();
            mString = newbuf;
        }
#endif // WIN32
    }
    
    /**
    * Reset the object. The string is emptied.
    *
    * @param from The value to assign to this object.
    *
    * @return A reference to this object.
    */
    void Reset()
    {
        if (MustErase() && !mString.empty())
        {
            memset(&mString[0], 
                0, 
                mString.size() * sizeof(charT));
        }
        mString.erase();
    }
    
    // ACCESS
    const std::basic_string<charT>& GetValue() const 
    {
        return mString;
    }
    // INQUIRY

    /**
    * Add a value
    *
    * Usually called from within "<<" operators
    *
    * @param from The value to add to this object.
    *
    * @return A reference to this object.
    */
    template <class defaultT> 
        WarCollector<charT>& Add(const defaultT& from) 
    {
        mString += from;
        return *this;
    }

    /**
    * Add an integer value
    *
    * Usually called from within "<<" operators
    *
    * @param from The value to add to this object.
    *
    * @return A reference to this object.
    */
    template <class numT> 
        WarCollector<charT>& AddInt(const numT& from) 
    {
        WarItoa<charT, numT>(mString, from);
        return *this;
    }

    /**
    * Add a newline
    *
    * Usually called from within "<<" operators
    *
    * @param from The value to add to this object.
    *
    * @return A reference to this object.
    */
    virtual WarCollector<charT>& AddEndln(const war_endl_e& from) 
    {
        Add((char)from);
        return *this;
    }

     /**
    * Add an error
    *
    * Usually called from within "<<" operators
    *
    * @param Ignored in the default implementation.
    *
    * @return A reference to this object.
    */
    virtual WarCollector<charT>& AddError(const WarError& from) 
    {
        *this << from.Explain();
        return *this;
    }

    inline bool MustErase() const
    {
        return mSecurityMode > SM_NONE;
    }
    
    std::basic_string<charT> mString;
private:
    SecurityModeE mSecurityMode; // Erase buffer when finished, used for password buffers
};

/* INLINE METHODS */

/* EXTERNAL REFERENCES */

/** */
WarCollector<char>& operator << (WarCollector<char>& c, 
                           const struct in_addr& val);

/** */
WarCollector<char>& operator << (WarCollector<char>& c, 
                           const char val);

/** */
WarCollector<wchar_t>& operator << (WarCollector<wchar_t>& c, 
                           const wchar_t val);

/** */
WarCollector<char>& operator << (WarCollector<char>& c, 
                           const wchar_t val);

WarCollector<wchar_t>& operator << (WarCollector<wchar_t>& c, 
                           const char val);

/** */
template <class charT>
WarCollector<charT>& operator << (WarCollector<charT>& c, 
                           const int val)
{
    return c.AddInt(val);
}
/** */
template <class charT>
WarCollector<charT>& operator << (WarCollector<charT>& c, 
                           const war_int64_t& val)
{
    return c.AddInt(val);
}

/** */
template <class charT>
WarCollector<charT>& operator << (WarCollector<charT>& c, 
                           const unsigned int val)
{
    return c.AddInt(val);
}

/** */
template <class charT>
WarCollector<charT>& operator << (WarCollector<charT>& c, 
                           const war_uint64_t& val)
{
    return c.AddInt(val);
}

/** */
WarCollector<char>& operator << (WarCollector<char>& c, 
                           const war_ccstr_t val);

/** */
WarCollector<char>& operator << (WarCollector<char>& c, 
                           const war_cwstr_t val);

/** */
WarCollector<wchar_t>& operator << (WarCollector<wchar_t>& c, 
                           const war_ccstr_t val);

/** */
WarCollector<wchar_t>& operator << (WarCollector<wchar_t>& c, 
                           const war_cwstr_t val);

/** */
WarCollector<char>& operator << (WarCollector<char>& c, 
                           const std::string& val);

/** */
WarCollector<char>& operator << (WarCollector<char>& c, 
                           const std::wstring& val);

/** */
WarCollector<wchar_t>& operator << (WarCollector<wchar_t>& c, 
                           const std::string& val);

/** */
WarCollector<wchar_t>& operator << (WarCollector<wchar_t>& c, 
                           const std::wstring& val);

/** */
template <class charT, class collT>
WarCollector<charT>& operator << (WarCollector<charT>& c, 
                           const WarCollector<collT>& val)
{
    c << val.GetValue();
    return c;
}

template <class charT>
WarCollector<charT>& operator << (WarCollector<charT>& c, 
                           const war_endl_e& val)
{
    return c.AddEndln(val);
}

/** */
template <class charT>
WarCollector<charT>& operator << (WarCollector<charT>& c, 
                           const WarError& val)
{
    return c.AddError(val);
}


#endif /* __cplusplus */

/****************** END C++ spesific ******************/

#endif  /* _WAR_COLLECTOR_H_ */

